
#include <iostream.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "grfhack.h"

// CLASS GrfFile
// Encapsulation is our friend =8*)
// If the outside world wants to see something, THEY HAVE TO ASK FIRST!!!!
// Plus it puts all the globals into one nice neat place
// Someday all errors will be handled using exceptions.  Ahhhh what a
// wonderful thought!

GrfFile::GrfFile()
{
	ddTable = 0;
	namTable = 0;
}

GrfFile::~GrfFile()
{
	if (grfFile)
		fclose(grfFile);
// I'm letting Windows de-allocate the memory for now
}

///////////////////////////////////////////////////////////////////////////////
// GrfFile::InitGrf                                                           //
// Given a filename, allocates and reads tables in GRF file                  //
//                                                                           //
// Input:  A pointer to the filename of a GRF file                           //
// Output: None                                                              //
// Return: Boolean value: true if successful, false if not                   //
///////////////////////////////////////////////////////////////////////////////
bool GrfFile::InitGrf(char * grfFileName)
{
	unsigned long temp;
//// Open Files ///////////////////////////////////////////////////////////////
	grfFile = fopen(grfFileName,"rb");
	if (!grfFile)
		return false;

//// Read and Verify Header ///////////////////////////////////////////////////
	fseek(grfFile, 0, SEEK_SET);
	temp = fread(&header, sizeof(grfHeader), 1, grfFile);
	if (!temp)
		return false;

	if (header.grfSig != 0x0ACEBABE)
		return false;

//// Allocate Memory for DDTable and Read /////////////////////////////////////
	ddTable = new ddEntry [header.numDDEs];
	if (!ddTable)
		return false;

	for(temp = 0; temp < header.numDDEs; temp++)
	{
		fread(&ddTable[temp], sizeof(ddEntry), 1, grfFile);
	}

//// Seek to NamTable, Allocate Memory and Read ///////////////////////////////
	fseek(grfFile, header.offNamTab, SEEK_SET);

	namTable = new ntEntry [header.numNTEs];
	if (!namTable)
		return false;

	for(temp = 0; temp < header.numNTEs; temp++)
	{
		fread(&namTable[temp], sizeof(ntEntry)-sizeof(char *), 1, grfFile);
		namTable[temp].string = new char [ namTable[temp].strLength ];
		if (!namTable[temp].string)
			return false;

		fread(namTable[temp].string, namTable[temp].strLength, 1, grfFile);
	}

//// Look at me I make cool comments I must be cool ;) ////////////////////////	
	ntNums = new unsigned long [header.numDDEs];
	for(temp = 0; temp < header.numDDEs; temp++)
	{
		unsigned long dataTag;

		dataTag = ddTable[temp].ddeTag;

		for( unsigned long i=0; i < header.numNTEs; i++ )
			if( namTable[i].nteTag == dataTag )
				break;

		ntNums[temp] = i;
	}

	return true;
}

///////////////////////////////////////////////////////////////////////////////
// GrfFile::ReadGrfBlock                                                      //
// Retrieve block of data from GRF file                                      //
//                                                                           //
// Input:  Block number                                                      //
// Output: Pointer to name of block                                          //
//         Pointer to actual data                                            //
//         Data Length                                                       //
// Return: Boolean true if successful, false if not                          //
///////////////////////////////////////////////////////////////////////////////
bool GrfFile::ReadGrfBlock (unsigned long blockNum, char * &name, char * &data,
							unsigned long & datalen)
{
// Allocate and prepare the name //////////////////////////////////////////////
	unsigned long NTNum = GetNTNum(blockNum);

	if( NTNum == header.numNTEs )
		return false;

	name = new char[namTable[NTNum].strLength];
	if (!name)
	{
		return false;
	}
	
	strcpy(name, namTable[NTNum].string);

// Allocate room for the data, read and return the pointer ////////////////////
	
	data = new char[ ddTable[blockNum].dataLength ];
	if (!data)
	{
		return false;
	}

	fseek(grfFile, ddTable[blockNum].dataOffset, SEEK_SET);
	fread(data, ddTable[blockNum].dataLength, 1, grfFile);

	datalen = ddTable[blockNum].dataLength;

	return true;
}



///////////////////////////////////////////////////////////////////////////////
// GrfFile::NumDataBlocks                                                    //
// Returns number of data blocks in the file                                 //
//                                                                           //
// Input:  None                                                              //
// Output: None                                                              //
// Return: Number of data blocks                                             //
///////////////////////////////////////////////////////////////////////////////
unsigned long GrfFile::NumDataBlocks()
{
	return header.numDDEs;
}

///////////////////////////////////////////////////////////////////////////////
// GrfFile::NumDataBlocks                                                    //
// Returns number of name blocks in the file???                              //
//                                                                           //
// Input:  None                                                              //
// Output: None                                                              //
// Return: Number of name blocks                                             //
///////////////////////////////////////////////////////////////////////////////
unsigned long GrfFile::NumNameBlocks()
{
	return header.numNTEs;
}

///////////////////////////////////////////////////////////////////////////////
// GrfFile::GetNTNum                                                         //
// Returns NameTable Entry number for given block                            //
//                                                                           //
// Input:  Data Directory block number                                       //
// Output: None                                                              //
// Return: NameTable Entry number, or header.numNTEs if there is no name     //
///////////////////////////////////////////////////////////////////////////////
unsigned long GrfFile::GetNTNum(unsigned long block)
{
	return ntNums[block];
}
///////////////////////////////////////////////////////////////////////////////
// GrfFile::GetBlockName                                                     //
// Retrieve name of name table entry                                         //
//                                                                           //
// Input:  Block Number                                                      //
//         Pointer to string (unallocated preferably)                        //
// Output: Valid string                                                      //
// Return: Boolean true if successful, false if not                          //
///////////////////////////////////////////////////////////////////////////////
bool GrfFile::GetBlockName(unsigned long namBlock, char * &name)
{
	unsigned long i = GetNTNum(namBlock);

	if( i == header.numNTEs )
		return false;

	name = new char[namTable[i].strLength];
	if (!name)
		return false;
	
	strcpy(name, namTable[i].string);
	return true;
}

///////////////////////////////////////////////////////////////////////////////
// GrfFile::GetDDE                                                           //
// Retrieve Data Directory Entry Header                                      //
//                                                                           //
// Input:  Block Number                                                      //
// Output: None                                                              //
// Return: Pointer to data directory entry structure                         //
///////////////////////////////////////////////////////////////////////////////
ddEntry * GrfFile::GetDDE(unsigned long block)
{
	ddEntry * dde;
	dde = new ddEntry;
	dde->ddeTag = ddTable[block].ddeTag;
	dde->unknown1 = ddTable[block].unknown1;
	dde->dataOffset = ddTable[block].dataOffset;
	dde->dataLength = ddTable[block].dataLength;
	dde->unknown2 = ddTable[block].unknown2;
	dde->unknown3 = ddTable[block].unknown3;
	dde->dataType = ddTable[block].dataType;
	dde->count = ddTable[block].count;
	dde->unknown4 = ddTable[block].unknown4;
	return dde;
}
