/* Version history:
   0.1  -- initial release
   0.15	-- 01/06/99
		   loads and displays extra WBH information
		   clearer error message if no file was specified on the command line
		   increased data buffer size (better performance?)
		   easier block browsing
   0.20 -- 01/10/99		   
		   mass extraction of data
	 	   auto extract to WAV


  Future plans: Decoding ADPCM Wavs
				Full Error Checking
				Recognition of all header data
				Search feature
				Gui interface
				Gui Browser
				Modification/Creation of TPA's

  Known problems:	Error checking is sketchy
					scanf sucks and _getch knows it... they fight alot :P
					display of extra header info doesn't line up right all the time.
					WAV files will not play in built in Win95/8 Players */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <conio.h>
#include <string.h>
#include "tpahack.cpp"

void displayTitle();
void displayByeBye();
char getMenuChoice();
void ViewInfo();
void ViewBlock();
void ExtractBlock();
void MassExtract();
void BlockInfo(unsigned long);
void WriteWaveHeader(unsigned long, FILE *);

static FILE *tpafile;
static struct TPAFILE thisTPA;

int main(int argc, char **argv)
{
	displayTitle();

	if (argc > 1)
		tpafile = fopen(argv[1], "rb");

	if (argc <= 1 || !is_good_tpa(tpafile))
	{
		fprintf(stderr, "please specify a valid tpa filename on the command line\n");
		_getch();
		exit(1);
	}
	
	printf("Loading TPA Data ... \n");
	if ( int ec = load_tpa_data( tpafile, &thisTPA ) )
	{
		fprintf(stderr, "there was an error number %d while reading the tpa file",
			ec);
		_getch();
		exit(ec);
	}

	char c;
	while( (c = getMenuChoice()) != 'Q' )
	{
		switch(c)
		{
		case 'V':
			ViewInfo();
			break;
		case 'B':
			ViewBlock();
			break;
		case 'E':
			ExtractBlock();
			break;
		case 'M':
			MassExtract();
			break;
		default:
			;
		}
	}

	displayByeBye();
	_fcloseall( );
	return 0;
}

void displayTitle()
{
        printf("\nWelcome to TPAHack v0.20 by PenguiN42 and Iguanaman -- 1998-1999");
        printf("\nThis is an UNSUPPORTED ALPHA version.  Use at your own risk!!!");
        printf("\nNeither THS nor the authors of this program are responsible for");
        printf("\nany problems this program may cause!  It hasn't broken our");
        printf("\nmachines, but you never know with C/C++\n");
}

void displayByeBye()
{
}

char getMenuChoice()
{
	printf("\nTPAHack Menu\n------------\n");
	printf("V) View TPA File Info\n");
	printf("B) View Block Data\n");
	printf("E) Extract Block to File\n");
	printf("M) MASS EXTRACTION!\n");
	printf("Q) Quit\n");
	printf("Enter Choice: ");

	char c = toupper(_getche());

	printf("\n");

	return c;
}

void ViewInfo()
{
	int i;

	printf("\n\n TPA File info\n");

	printf("TAG: ");
	for (i=0; i<4; i++)
		printf("%02X ", thisTPA.tag[i]);
	printf("\nNumber of Blocks: %lu", thisTPA.numWBBlocks);
	printf("\nNumber of Table Entries: %lu", thisTPA.numWTEs);
	printf("\nFile Size: %lu", thisTPA.fileSize);
	printf("\nExtra Info: ");
	for (i=0; i<8; i++)
		printf("%02X ", thisTPA.data1[i]);
	printf("\nPress a key...");

	char c = _getch();

	printf("\n\n");
}

void ViewBlock()
{
	unsigned long num=0;

	printf("Enter block number: ");
	
	_getch(); // for some reason the "b" stays in the buffer
	scanf("%u", &num);

	if (num >= 0 && num < thisTPA.numWTEs)
	{
		int keepGoing = 1;
		while(keepGoing)
		{
			BlockInfo(num);
			printf("\n(P)revious, (N)ext, or any other key to finish.");
			switch (toupper( _getche() ))
			{
			case 'P':
				if (num <= 1) num = 0;
				else num--;
				break;
			case 'N':
				num++;
				if (num >= thisTPA.numWTEs) num = thisTPA.numWTEs - 1;
				break;
			default:
				keepGoing = 0;
			}
		}
	}
	else
		printf("\nValid Blocks are between 0 and %lu", thisTPA.numWTEs - 1);

	printf("\n\n");
}

void ExtractBlock()
{
	unsigned long num=0;

	printf("Enter block number: ");
	
	_getch(); // for some reason the "e" stays in the buffer
	scanf("%u", &num);

	if (num >= 0 && num < thisTPA.numWTEs)
	{
		BlockInfo(num);
		printf("\nThis block? Y/N: ");
		char c = toupper(_getche());

		if (c == 'Y')
		{
			char filename[256];
			FILE *outfile;
			printf("\nEnter filename: ");
			_getch();
			scanf("%s", filename);

			outfile = fopen(filename, "wb");
			
			if (!outfile)
			{
				printf("\nFailure opening file!");
				return;
			}

			printf("\nExtract Data in (R)aw or (W)av format?");
			if (toupper(_getche() ) == 'W')
				WriteWaveHeader(num, outfile);
						
			extract_WBBlock(tpafile, &thisTPA, num, outfile);

			fclose(outfile);
		}
	}
	else
		printf("\nValid Blocks are between 0 and %lu", thisTPA.numWTEs - 1);

	printf("\n\n");
}

void MassExtract()
{
	unsigned long count, startNum, endNum;
	int	 temp;
	char fileName[256];
	char directory[256];

	char tempName[16];
	char rawwav;
	FILE *outfile;

	printf("\nExtract Data in (R)aw or (W)av format?");
	rawwav = toupper( _getche() );

	printf("\nEnter directory to which extraction should take place:");
	_getch();
	scanf("%s", directory);

	printf("\nValid Blocks are between 0 and %lu", thisTPA.numWTEs - 1);

	printf("\nEnter block number at which to start extraction:");
	_getch();
	scanf("%u", &startNum);
	
	printf("\nEnter block number at which to end extraction:");
	_getch();
	scanf("%u", &endNum);

	if (endNum > (thisTPA.numWTEs-1) )
	{
		printf("\nInvalid Ending Block Number");
		return;
	}

	if (startNum > endNum)
	{
		printf("\nInvalid Range");
		return;
	}

	temp = strlen(directory);

	if (directory[temp-1] != '\\')
	{
		directory[temp] = '\\';
		directory[temp+1] = '\0';
	}
		
	for(count = startNum; count <= endNum; count++)
	{
		itoa(count, tempName, 10);

		strcpy(fileName, directory);
		strcat(fileName, tempName);
			
		if (rawwav = 'W')
			strcat(fileName, ".wav");
		else
			strcat(fileName, ".raw");

		printf("\n%s", fileName);

			
		outfile = fopen(fileName, "wb");

		if (!outfile)
		{
			printf("\nFailure opening file!");
			return;
		}

		if (rawwav == 'W')
		{
			WriteWaveHeader(count, outfile);
		}

		extract_WBBlock(tpafile, &thisTPA, count, outfile);
					
		fclose(outfile);
			
	}

}

void BlockInfo(unsigned long num)
{
	int i, j;

	printf("\n\n Info on Block #%lu", num);
	printf("\nOffset: %lu", thisTPA.WBTable[num]->blockOffset);
	printf("\nLength: %lu", thisTPA.WBTable[num]->blockLength);
	printf("\nHeader Length: %lu", thisTPA.WBTable[num]->WBHLength);
	printf("\nData Length: %lu", thisTPA.WBTable[num]->dataLength);
	printf("\nDATA1: ");
	for (i=0; i<4; i++)
		printf("%02X ", thisTPA.WBTable[num]->data1[i]);
	printf("\nDATA2: ");
	for (i=0; i<8; i++)
		printf("%02X ", thisTPA.WBTable[num]->data2[i]);
	if (thisTPA.WBTable[num]->adpBlkLen != 0)
		printf("\nADPCM Block Length: %lu", thisTPA.WBTable[num]->adpBlkLen);
	if (thisTPA.WBTable[num]->adpBlkLen != 0)
	{
		printf("\nUncompressed Data Length: %lu", thisTPA.WBTable[num]->unCompLen);
	}
	printf("\nSample Rate: %lu", thisTPA.WBTable[num]->sampRate);
	printf("  Bits/Sample: %u", thisTPA.WBTable[num]->numBits);
	printf("  Channels: %u", thisTPA.WBTable[num]->numChans);
	printf("  Format: ");
	switch (thisTPA.WBTable[num]->codeFormat)
	{
	case 0:
		printf("Raw PCM");
		break;
	case 1:
		printf("ADPCM");
		break;
	default:
		printf("Other");
		break;
	};

	printf("\nDATA6: %u", thisTPA.WBTable[num]->data6);
	printf("\nDATA7: ");
	for (i=0; i<4; i++)
		printf("%02X ", thisTPA.WBTable[num]->data7[i]);

	int extraHeadLen = thisTPA.WBTable[num]->WBHLength - 0x024;
	if(extraHeadLen > 0)	// block has extra header data
	{
		int offset = 0;
		printf("\nBLOCK HAS EXTRA HEADER DATA... PRESS A KEY");
		_getch();
		printf("\n Extra Header Data: \n");
		for(i=0; i < (extraHeadLen / 8); i++)
		{
			printf("\t\t");
			for(j=0; j < 8; j++, offset++)
			{
				printf("%02X ", thisTPA.WBTable[num]->extraData[offset]);
			}
			offset -= 8;
			printf("\t");
			for(j=0; j < 8; j++, offset++)
			{
				char c = thisTPA.WBTable[num]->extraData[offset];
				if (c >= ' ' && c <= '~')
					printf("%c", c);
				else
					printf(".");
			}
			printf("\n");
		}
		printf("\t\t");
		for(j=0; j < extraHeadLen % 8; j++, offset++)
		{
			printf("%02X ", thisTPA.WBTable[num]->extraData[offset]);
		}
		offset -= extraHeadLen % 8;
		printf("\t");
		for(j=0; j < extraHeadLen % 8; j++, offset++)
		{
			char c = thisTPA.WBTable[num]->extraData[offset];
			if (c >= ' ' && c <= '~')
				printf("%c", c);
			else
				printf(".");
		}
		printf("\n");
	}

	printf("\n");
}

void WriteWaveHeader(unsigned long num, FILE *outfile)
{

	if (thisTPA.WBTable[num]->codeFormat == 0)
	{
		PCMWAVE head;
		head.riffTag = 0x46464952;	// RIFF
		head.length = thisTPA.WBTable[num]->dataLength + long(36);
		head.waveTag = 0x45564157;	// WAVE
		head.fmtTag = 0x20746D66;	// fmt
		head.fmtLen = 0x10;
		head.encodingTag = 0x1;
		head.numChans = thisTPA.WBTable[num]->numChans;
		head.sampRate = thisTPA.WBTable[num]->sampRate;
		head.bitsSamp = thisTPA.WBTable[num]->numBits;
		head.blockAlign = head.numChans * head.bitsSamp / 8;
		head.dataRate = head.sampRate * long(head.blockAlign);
		head.dataTag = 0x61746164;	// DATA
		head.dataLen = thisTPA.WBTable[num]->dataLength;
		fwrite(&head, sizeof(head), 1, outfile );
	}
	else if (thisTPA.WBTable[num]->codeFormat == 1)
	{
		ADPCMWAVE head;
		unsigned long sampPerBlk;
		
		sampPerBlk = thisTPA.WBTable[num]->adpBlkLen;
		sampPerBlk = sampPerBlk / 4;
		sampPerBlk = sampPerBlk / thisTPA.WBTable[num]->numChans;
		sampPerBlk = (sampPerBlk - 1) * 8;
		head.riffTag = 0x46464952;	// RIFF
		head.length = thisTPA.WBTable[num]->dataLength + long(52);
		head.waveTag = 0x45564157;	// WAVE
		head.fmtTag = 0x20746D66;	// fmt
		head.fmtLen = 0x14;
		head.encodingTag = 0x11;
		head.numChans = thisTPA.WBTable[num]->numChans;
		head.sampRate = thisTPA.WBTable[num]->sampRate;
		head.bitsSamp = thisTPA.WBTable[num]->numBits / 4;
		head.adpBlkLen = short(thisTPA.WBTable[num]->adpBlkLen);
		head.dataRate = head.sampRate * head.numChans * head.bitsSamp / 8;
		head.cbSize = 0x02;
		head.sampPerBlk = short(sampPerBlk);
		head.factTag = 0x74636166;	// fact
		head.factBits = head.bitsSamp;
		head.unCompLen = thisTPA.WBTable[num]->unCompLen;
		head.dataTag = 0x61746164;	// DATA
		head.dataLen = thisTPA.WBTable[num]->dataLength;
		fwrite(&head, sizeof(head), 1, outfile );
	}

	

}
